__all__ = ['ThingDropped', 'FieldValidationError', 'ItemExtractError', 'QueenError']


class ThingDropped(Exception):
    """Raise when pipeline dropped one thing"""


class FieldValidationError(Exception):
    pass


class ItemExtractError(Exception):
    """For extract item"""


class QueenError(Exception):
    """For "queen" module"""



"""The thing`s usage is simple, can be created by ant, processed by ants and pipelines"""
from typing import Any, Optional, Iterator, Tuple, Dict, Type, Union, List, DefaultDict, AnyStr, IO, Callable
from collections.abc import MutableMapping
import json
import abc
from collections import defaultdict
import logging
import re

from aiohttp import ClientRequest, ClientResponse
from lxml import html
import jpath

from .exceptions import FieldValidationError, ItemExtractError


class Request(ClientRequest):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # store data obj
        self.data = kwargs.get('data', None)  # type: Union[AnyStr, dict, IO, None]


class Response(ClientResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._text = None
        self._html_element = None
        self._json = None

    def get_text(self, encoding: Optional[str]=None, errors: str='strict') -> str:
        if self._content is None:
            raise ValueError('Read stream first')
        if self._text is None:
            if encoding is None:
                encoding = self._get_encoding()
            self._text = self._content.decode(encoding, errors=errors)
        return self._text

    @property
    def simple_text(self) -> str:
        return self.get_text()

    def get_json(self, encoding: Optional[str] = None, decode_errors: str='strict',
                 loads: Callable=json.loads):
        if self._text is None:
            text = self.get_text(encoding=encoding, errors=decode_errors)
        else:
            text = self._text
        if self._json is None:
            self._json = loads(text)
        return self._json

    @property
    def simple_json(self) -> Any:
        return self.get_json()

    @property
    def html_element(self) -> html.HtmlElement:
        if self._html_element is None:
            self._html_element = html.fromstring(self.simple_text)
        return self._html_element


class IntField:
    _type = int
    storage_name = ''
    __shadow_name_prefix = '__field#'

    def __init__(self, null: bool=False, default: Any=None):
        """
        "null" is True means this field can be ignore when value have not been set in validation,
        "default" is None means this field have no default value
        """
        self.null = null
        self.default = default

    def __set__(self, instance: 'Item', value: Any) -> None:
        setattr(instance, self.storage_name, value)

    def __get__(self, instance: 'Item', owner: Type['Item']) -> Any:
        try:
            return getattr(instance, self.storage_name)
        except AttributeError as e:
            raise AttributeError(
                '\'{:s}\' object has no attribute \'{:s}\''.format(instance.__class__.__name__,
                                                                   self.get_name_from_shadow(self.storage_name))) from e

    def __delete__(self, instance):
        delattr(instance, self.storage_name)

    def validate(self, value: Any) -> Any:
        """:raise FieldValidationError"""
        try:
            return self._type(value)
        except (ValueError, TypeError) as e:
            raise FieldValidationError(str(e)) from e

    @classmethod
    def make_shadow_name(cls, name: str) -> str:
        if name == cls.__shadow_name_prefix:
            raise AttributeError('This name: {:s} has been used internally'.format(cls.__shadow_name_prefix))
        return cls.__shadow_name_prefix + name

    @classmethod
    def is_shadow_name(cls, name: str) -> bool:
        return cls.__shadow_name_prefix in name

    @classmethod
    def get_name_from_shadow(cls, name: str) -> str:
        return name.replace(cls.__shadow_name_prefix, '')


class FloatField(IntField):
    _type = float


class StringField(IntField):
    _type = str


class BytesField(IntField):
    _type = bytes


class ItemMeta(abc.ABCMeta):
    def __init__(cls, name: str, bases: Tuple[type, ...], attr_dict: Dict[str, Any]):
        super().__init__(name, bases, attr_dict)
        for k, v in attr_dict.items():
            if isinstance(v, IntField):
                v.storage_name = IntField.make_shadow_name(k)


class Item(MutableMapping, metaclass=ItemMeta):
    _is_validating = False

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        # default field value
        class_dict = self.__class__.__dict__
        for k, obj in class_dict.items():
            if isinstance(obj, IntField):
                if obj.default is not None and k not in self:
                    setattr(self, k, obj.default)

    def __setattr__(self, key: str, value: Any) -> None:
        if isinstance(key, str):
            super().__setattr__(key, value)
        else:
            raise AttributeError('attribute name must be string, not \'{:s}\''.format(key.__class__.__name__))

    def __setitem__(self, key: str, value: Any) -> None:
        try:
            setattr(self, key, value)
        except TypeError as e:
            raise KeyError(str(e)) from e

    def __getitem__(self, key: str) -> Any:
        try:
            return getattr(self, key)
        except AttributeError as e:
            raise KeyError(str(e)) from e

    def __delitem__(self, key: str) -> None:
        try:
            delattr(self, key)
        except AttributeError as e:
            raise KeyError(str(e)) from e

    def __len__(self) -> int:
        return len(self.__dict__)

    def __iter__(self) -> Iterator[str]:
        keys = []
        for k in self.__dict__:
            if IntField.is_shadow_name(k):
                keys.append(IntField.get_name_from_shadow(k))
            else:
                keys.append(k)
        return iter(keys)

    def validate(self):
        """Validate item`s type.
        Get descriptors reference method from __class__.__dict__"""
        class_dict = self.__class__.__dict__
        for k, obj in class_dict.items():
            if isinstance(obj, IntField):
                if k in self:
                    setattr(self, k, class_dict[k].validate(getattr(self, k)))
                elif not obj.null:
                    raise FieldValidationError(
                        '\'{:s}.{:s}\' have no value yet'.format(self.__class__.__name__, k))

    def __repr__(self):
        return '{:s}: {:s}'.format(self.__class__.__name__, str(dict(self)))

    def __str__(self):
        return '{:s}'.format(self.__class__.__name__)


Things = Union[Request, Response, Item]


class CustomNoneType:
    pass


class ItemExtractor:
    extract_with_take_first = 'take_first'
    extract_with_join_all = 'join_all'
    extract_with_do_nothing = 'do_nothing'

    def __init__(self, item_class: Type[Item]):
        self.item_class = item_class
        self.logger = logging.getLogger(self.__class__.__name__)
        self.path = defaultdict(list)  # type: DefaultDict[str, List[Tuple[str, str, str]]]

    def add_xpath(self, key: str, xpath: str, extract_type=extract_with_take_first):
        self.path[key].append(('xpath', xpath, extract_type))

    def add_regex(self, key: str, pattern: str, extract_type=extract_with_take_first):
        self.path[key].append(('regex', pattern, extract_type))

    def add_jpath(self, key: str, jpath, extract_type=extract_with_take_first):
        self.path[key].append(('jpath', jpath, extract_type))

    @staticmethod
    def extract_value(_type: str, pattern: str, data: Any, extract_type=extract_with_take_first) -> Any:
        if _type == 'xpath':
            extract_value = data.xpath(pattern)
        elif _type == 'regex':
            extract_value = re.findall(pattern, data)
        elif _type == 'jpath':
            extract_value = jpath.get_all(pattern, data)
        else:
            raise ValueError('The type: {:s} not support'.format(_type))
        # handle by extract type
        if extract_type == ItemExtractor.extract_with_take_first:
            extract_value = extract_value[0]
        elif extract_type == ItemExtractor.extract_with_join_all:
            extract_value = list(filter(lambda x: isinstance(x, str), extract_value))  # join string only
            extract_value = ''.join(extract_value)
        return extract_value

    def extract(self, response: Response) -> Item:
        """Extract item from response by path with xpath, jpath or re."""
        self.logger.debug('Extract item: {:s} with path: {:s}'.format(self.item_class.__name__, str(self.path)))
        item = self.item_class()
        for key, all_xpath in self.path.items():
            value = CustomNoneType()  # be different with "None" obj ("null" in json)
            for path_type, path, extract_type in all_xpath:
                # get data by search type
                if path_type == 'xpath':
                    data = response.html_element
                elif path_type == 'regex':
                    data = response.simple_text
                else:
                    data = response.simple_json
                try:
                    extract_value = ItemExtractor.extract_value(path_type, path, data, extract_type=extract_type)
                except IndexError:
                    continue
                # check multiple path`s result
                if not isinstance(value, CustomNoneType) and value != extract_value:
                    raise ItemExtractError(
                        'Match different result: {:s} and {:s} for key: {:s}'.format(value, extract_value, key))
                value = extract_value
            if not isinstance(value, CustomNoneType):
                item[key] = value
        return item


__all__ = ['Request', 'Response', 'Item', 'ItemExtractor', 'Things'] + [var for var in vars().keys() if 'Field' in var]



from typing import Optional, List, Coroutine, Union, Dict, Callable, AnyStr, IO, DefaultDict
import abc
import itertools
import logging
import time
from collections import defaultdict

import aiohttp
from aiohttp.client import DEFAULT_TIMEOUT
from aiohttp import ClientSession
import async_timeout
from yarl import URL
from tenacity import retry
from tenacity.retry import retry_if_result, retry_if_exception_type
from tenacity.wait import wait_fixed
from tenacity.stop import stop_after_attempt

from .pipelines import Pipeline
from .things import Request, Response, Item, Things
from . import queen


__all__ = ['Ant']


class Ant(abc.ABC):
    response_pipelines = []  # type: List[Pipeline]
    request_pipelines = []  # type: List[Pipeline]
    item_pipelines = []  # type: List[Pipeline]
    request_timeout = DEFAULT_TIMEOUT
    request_retries = 3
    request_retry_delay = 5
    request_proxy = None  # type: Optional[str]
    request_max_redirects = 10
    request_allow_redirects = True
    response_in_stream = False

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.sessions = {}  # type: Dict[str, ClientSession]
        # report var
        self._reports = defaultdict(lambda: [0, 0])  # type: DefaultDict[str, List[int, int]]
        self._drop_reports = defaultdict(lambda: [0, 0])  # type: DefaultDict[str, List[int, int]]
        self._start_time = time.time()
        self._last_time = self._start_time
        self._report_slot = 60  # report once after one minute by default

    async def request(self, url: Union[str, URL], method='GET', params: Optional[dict]=None,
                      headers: Optional[dict]=None, cookies: Optional[dict]=None,
                      data: Optional[Union[AnyStr, Dict, IO]]=None,
                      ) -> Response:
        if not isinstance(url, URL):
            url = URL(url)
        req = Request(method, url, params=params, headers=headers, cookies=cookies, data=data)
        req = await self._handle_thing_with_pipelines(req, self.request_pipelines, timeout=self.request_timeout)
        self.report(req)

        request_function = queen.timeout_wrapper(self._request, timeout=self.request_timeout)
        retries = self.request_retries
        if retries > 0:
            res = await self.make_retry_decorator(retries, self.request_retry_delay)(request_function)(req)
        else:
            res = await request_function(req)

        res = await self._handle_thing_with_pipelines(res, self.response_pipelines, timeout=self.request_timeout)
        self.report(res)
        return res

    async def collect(self, item: Item) -> None:
        self.logger.debug('Collect item: ' + str(item))
        await self._handle_thing_with_pipelines(item, self.item_pipelines)
        self.report(item)

    async def open(self) -> None:
        self.logger.info('Opening')
        for pipeline in itertools.chain(self.item_pipelines, self.response_pipelines, self.request_pipelines):
            try:
                obj = pipeline.on_spider_open()
                if isinstance(obj, Coroutine):
                    await obj
            except Exception as e:
                self.logger.exception('Open pipelines with ' + e.__class__.__name__)

    async def close(self) -> None:
        for pipeline in itertools.chain(self.item_pipelines, self.response_pipelines, self.request_pipelines):
            try:
                obj = pipeline.on_spider_close()
                if isinstance(obj, Coroutine):
                    await obj
            except Exception as e:
                self.logger.exception('Close pipelines with ' + e.__class__.__name__)
        # close cached sessions
        for session in self.sessions.values():
            await session.close()
        self.logger.info('Closed')

    @abc.abstractmethod
    async def run(self) -> None:
        """App custom entrance"""

    async def main(self) -> None:
        await self.open()
        try:
            await self.run()
        except Exception as e:
            self.logger.exception('Run ant run`s coroutine with ' + e.__class__.__name__)
        # wait scheduled coroutines before wait "close" method
        await queen.wait_scheduled_coroutines()
        await self.close()
        await queen.wait_scheduled_coroutines()
        # total report
        for name, counts in self._reports.items():
            self.logger.info('Get {:d} {:s} in total'.format(counts[1], name))
        for name, counts in self._drop_reports.items():
            self.logger.info('Drop {:d} {:s} in total'.format(counts[1], name))
        self.logger.info('Run {:s} in {:f} seconds'.format(self.__class__.__name__, time.time() - self._start_time))

    @staticmethod
    def make_retry_decorator(retries: int, delay: float) -> Callable[[Callable], Callable]:
        return retry(wait=wait_fixed(delay),
                     retry=(retry_if_result(lambda res: res.status >= 500) | retry_if_exception_type()),
                     stop=stop_after_attempt(retries + 1))

    async def _handle_thing_with_pipelines(self, thing: Things, pipelines: List[Pipeline],
                                           timeout=DEFAULT_TIMEOUT) -> Things:
        """Process thing one by one, break the process chain when get "None" or exception
        :raise ThingDropped"""
        self.logger.debug('Process thing: ' + str(thing))
        raw_thing = thing
        for pipeline in pipelines:
            thing = pipeline.process(thing)
            if isinstance(thing, Coroutine):
                with async_timeout.timeout(timeout):
                    thing = await thing
            if isinstance(thing, Exception):
                self.report(raw_thing, dropped=True)
                raise thing
        return thing

    async def _request(self, req: Request) -> Response:
        # cookies in headers, params in url
        kwargs = dict(method=req.method, url=req.url, headers=req.headers, data=req.data)
        kwargs['proxy'] = self.request_proxy
        kwargs['max_redirects'] = self.request_max_redirects
        kwargs['allow_redirects'] = self.request_allow_redirects

        # proxy auth not work when one session with many requests, add auth header to fix it
        proxy = None if kwargs['proxy'] is None else URL(kwargs['proxy'])
        if proxy is not None and proxy.user is not None:
            if kwargs['headers'] is None:
                kwargs['headers'] = dict()
            kwargs['headers'][aiohttp.hdrs.PROXY_AUTHORIZATION] = aiohttp.BasicAuth.from_url(proxy).encode()

        host = req.url.host
        if host not in self.sessions:
            session = aiohttp.ClientSession(response_class=Response)
            self.sessions[host] = session
        else:
            session = self.sessions[host]

        response = await session._request(**kwargs)
        if not self.response_in_stream:
            await response.read()
            response.close()
            await response.wait_for_close()
        return response

    def report(self, thing: Things, dropped: bool=False) -> None:
        now_time = time.time()
        if now_time - self._last_time > self._report_slot:
            self._last_time = now_time
            for name, counts in self._reports.items():
                count = counts[1] - counts[0]
                counts[0] = counts[1]
                self.logger.info(
                    'Get {:d} {:s} in total with {:d}/{:d}s rate'.format(
                        counts[1], name, count, self._report_slot))
            for name, counts in self._drop_reports.items():
                count = counts[1] - counts[0]
                counts[0] = counts[1]
                self.logger.info(
                    'Drop {:d} {:s} in total with {:d}/{:d} rate'.format(
                        counts[1], name, count, self._report_slot))
        report_type = thing.__class__.__name__
        if dropped:
            reports = self._drop_reports
        else:
            reports = self._reports
        counts = reports[report_type]
        counts[1] += 1



"""
Ant`s queen, provide coroutines manage function
"""
import typing
import logging
import asyncio
from asyncio.queues import Queue, QueueEmpty
from itertools import islice

import async_timeout
from aiohttp.client import DEFAULT_TIMEOUT

from .exceptions import QueenError

logger = logging.getLogger(__name__)
__queue = None  # type: typing.Optional[Queue]
__done_queue = None  # type: typing.Optional[Queue]
__running_count = 0
__timeout = DEFAULT_TIMEOUT  # in seconds
__concurrent_limit = 30
# The Queue obj and "schedule_coroutine" function should use the same event loop
__loop = None  # type: typing.Optional[asyncio.AbstractEventLoop]


__all__ = ['init_loop', 'reset_concurrent_limit', 'get_loop', 'timeout_wrapper', 'schedule_coroutine',
           'schedule_coroutines', 'as_completed', 'wait_scheduled_coroutines']


def init_loop(loop: typing.Optional[asyncio.AbstractEventLoop]=None) -> None:
    """Set event loop and coroutines queue for "schedule_coroutine" function"""
    global __queue, __done_queue, __loop
    if loop is None:
        loop = asyncio.get_event_loop()

    if __loop is not None:
        if __running_count > 0 or __done_queue.qsize() > 0 or __queue.qsize() > 0:
            raise QueenError('Replace Event loop with uncleaned coroutines')

    __loop = loop
    __queue = Queue(loop=__loop)
    __done_queue = Queue(loop=__loop)


def get_loop() -> asyncio.AbstractEventLoop:
    if __loop is None:
        init_loop()
    return __loop


def reset_concurrent_limit(limit: typing.Optional[int]) -> None:
    global __concurrent_limit
    if limit is not None:
        __concurrent_limit = limit


def timeout_wrapper(coro_or_func: typing.Union[typing.Coroutine, typing.Callable], timeout: float=__timeout
                    ) -> typing.Union[typing.Coroutine, typing.Callable]:
    """Add timeout limit to coroutine or coroutine function"""

    async def wrapper(*args, **kwargs):
        with async_timeout.timeout(timeout):
            if asyncio.iscoroutinefunction(coro_or_func):
                return await coro_or_func(*args, **kwargs)
            else:
                return await coro_or_func

    if asyncio.iscoroutinefunction(coro_or_func):
        return wrapper
    else:
        return wrapper()


def schedule_coroutine(coroutine: typing.Coroutine, timeout: float=__timeout) -> None:
    """Like "ensure_future", it schedule coroutine with concurrent limit even in "run_until_commplete" loop,
    "wait_scheduled_coroutines" coroutine must be called before process exit make sure all coroutine has been done"""
    if __loop is None or __queue is None or __done_queue is None:  # init Queue obj
        init_loop(loop=__loop)

    global __running_count

    def _done_callback(f):
        global __running_count

        exception = f.exception()
        if exception is not None:
            try:
                raise exception
            except exception.__class__:
                logger.exception(exception)

        __running_count -= 1
        __done_queue.put_nowait(f)
        try:
            if __running_count < __concurrent_limit:
                next_coroutine = __queue.get_nowait()
                __running_count += 1
                asyncio.ensure_future(next_coroutine, loop=__loop).add_done_callback(_done_callback)
        except QueueEmpty:
            pass
    if __running_count < __concurrent_limit:
        __running_count += 1
        asyncio.ensure_future(
            timeout_wrapper(coroutine, timeout=timeout), loop=__loop).add_done_callback(_done_callback)
    else:
        __queue.put_nowait(timeout_wrapper(coroutine, timeout=timeout))


def schedule_coroutines(coroutines: typing.Iterable, timeout: float=__timeout) -> None:
    for coroutine in coroutines:
        schedule_coroutine(coroutine, timeout=timeout)


async def wait_scheduled_coroutines():
    """Wait all coroutines schedule by "schedule_coroutine" function"""
    if __loop is None:
        return
    while __running_count > 0 or __done_queue.qsize() > 0:
        await __done_queue.get()


def as_completed(coroutines: typing.Iterable[typing.Coroutine], limit: int=__concurrent_limit, timeout: float=__timeout,
                 loop: typing.Optional[asyncio.AbstractEventLoop]=None
                 )-> typing.Generator[typing.Coroutine, None, None]:
    """Custom as_completed method provide coroutines concurrent limit,
    the "limit" is not shared with "schedule_coroutine" function"""
    coroutines = iter(coroutines)

    queue = Queue(loop=loop)
    todo = []

    def _done_callback(f):
        queue.put_nowait(f)
        todo.remove(f)
        try:
            nf = asyncio.ensure_future(next(coroutines))
            nf.add_done_callback(_done_callback)
            todo.append(nf)
        except StopIteration:
            pass

    async def _wait_for_one():
        f = await queue.get()
        return f.result()

    if limit <= 0:
        fs = {asyncio.ensure_future(timeout_wrapper(cor, timeout=timeout), loop=loop) for cor in coroutines}
    else:
        fs = {asyncio.ensure_future(timeout_wrapper(cor, timeout=timeout), loop=loop)
              for cor in islice(coroutines, 0, limit)}
    for f in fs:
        f.add_done_callback(_done_callback)
        todo.append(f)

    while len(todo) > 0 or queue.qsize() > 0:
        yield _wait_for_one()


async def as_completed_with_async(
        coroutines: typing.Iterable[typing.Coroutine],
        limit: int=__concurrent_limit, timeout: float=__timeout, ignore_exception: bool=True,
        )-> typing.AsyncGenerator[typing.Any, None]:
    """as_completed`s async version, catch and log exception inside"""
    for coro in as_completed(coroutines, limit=limit, timeout=timeout, loop=asyncio.get_event_loop()):
        try:
            yield await coro
        except Exception as e:
            if ignore_exception:
                logger.exception(e)
            else:
                raise e



from typing import Optional, List, Tuple, DefaultDict, Dict, Any, IO, Union
import logging
from collections import defaultdict
import json
import os
import datetime
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
import asyncio

import aiomysql
import aiosmtplib

from .things import Things, Response, Request, Item
from .exceptions import FieldValidationError, ThingDropped


class Pipeline:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    async def on_spider_open(self) -> None:
        """Call when ant open, method or coroutine"""

    async def on_spider_close(self) -> None:
        """Call when ant close, method or coroutine"""

    async def process(self, thing: Things) -> Union[Things, Exception]:
        """method or coroutine"""
        return thing


# Response pipelines
class ResponseFilterErrorPipeline(Pipeline):
    def process(self, thing: Response) -> Union[Things, Exception]:
        if thing.status >= 400:
            return ThingDropped('Respose status is {:d}'.format(thing.status))
        else:
            return thing


# Request pipelines
class RequestDuplicateFilterPipeline(Pipeline):
    def __init__(self):
        self.__request_urls = set()
        super().__init__()

    def process(self, thing: Request) -> Union[Things, Exception]:
        if thing.url in self.__request_urls:
            return ThingDropped('Request duplicate!')
        else:
            self.__request_urls.add(thing.url)
            return thing


class RequestUserAgentPipeline(Pipeline):
    user_agent = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_1) AppleWebKit/537.36 (KHTML, like Gecko)' \
                 'Chrome/62.0.3202.89 Safari/537.36 Name'

    def __init__(self, user_agent=user_agent):
        super().__init__()
        self.user_agent = user_agent

    def process(self, thing):
        if thing.headers is None:
            thing.headers = {}
        thing.headers['User-Agent'] = self.user_agent
        return thing


# Item pipelines
class ItemPrintPipeline(Pipeline):
    def process(self, thing: Item) -> Item:
        self.logger.info(thing.__repr__())
        return thing


class ItemValidatePipeline(Pipeline):
    def process(self, thing: Item) -> Union[Things, Exception]:
        try:
            thing.validate()
            return thing
        except FieldValidationError as e:
            return e


class ItemFieldReplacePipeline(Pipeline):
    def __init__(self, fields: List[str], excess_chars: Tuple[str]=('\r', '\n', '\t')):
        self.fields = fields
        self.excess_chars = excess_chars
        super().__init__()

    def process(self, thing: Item) -> Item:
        for field in self.fields:
            for char in self.excess_chars:
                if field in thing and isinstance(thing[field], str):
                    thing[field] = thing[field].replace(char, '')
        return thing


class ItemBaseJsonDumpPipeline(Pipeline):
    @staticmethod
    def _dump(file_path: str, data: Dict) -> None:
        with open(file_path, 'w') as f:
            json.dump(data, f)

    async def dump(self, file_path: str, data: Dict) -> None:
        await asyncio.get_event_loop().run_in_executor(None, self._dump, file_path, data)


class ItemJsonDumpPipeline(ItemBaseJsonDumpPipeline):
    def __init__(self, file_dir: str='.'):
        super().__init__()
        self.file_dir = file_dir
        self.data = defaultdict(list)  # type: DefaultDict[str, List[Dict]]

    def process(self, thing: Item) -> Item:
        self.data[thing.__class__.__name__].append(dict(thing))
        return thing

    async def on_spider_close(self) -> None:
        for file_name, data in self.data.items():
            await self.dump(os.path.join(self.file_dir, file_name + '.json'), data)


class ItemBaseMysqlPipeline(Pipeline):
    def __init__(self, host: str, port: int, user: str, password: str, database: str, table: str, charset: str='utf8'):
        super().__init__()
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.table = table
        self.charset = charset
        self.pool = None

    async def on_spider_open(self) -> None:
        self.pool = await aiomysql.create_pool(host=self.host, port=self.port, user=self.user, password=self.password,
                                               db=self.database, charset=self.charset, use_unicode=True)

    async def on_spider_close(self) -> None:
        self.pool.close()
        await self.pool.wait_closed()

    async def push_data(self, sql: str) -> None:
        """Run SQL without pulling data like "INSERT" and "UPDATE" command"""
        self.logger.debug('Executing SQL: ' + sql)
        async with self.pool.get() as conn:
            async with conn.cursor() as cur:
                await cur.execute(sql)
            await conn.commit()

    async def pull_data(self, sql: str) -> Tuple[Dict[str, Any]]:
        """Run SQL with pulling data like "SELECT" command"""
        self.logger.debug('Executing SQL: ' + sql)
        async with self.pool.get() as conn:
            async with conn.cursor(aiomysql.DictCursor) as cur:
                await cur.execute(sql)
                return await cur.fetchall()

    @staticmethod
    def convert_item_value(value: Any) -> str:
        """Parse value to str for making sql string, eg: False -> '0'"""
        if isinstance(value, bool):
            return '1' if value else '0'
        elif isinstance(value, str):
            value = value.replace('\"', '\\\"')
            return '"{:s}"'.format(value)
        elif isinstance(value, datetime.datetime):
            return '"{:s}"'.format(str(value))
        elif isinstance(value, datetime.date):
            return '"{:s}"'.format(str(value))
        elif value is None:
            return 'null'
        else:
            return str(value)


class ItemMysqlInsertPipeline(ItemBaseMysqlPipeline):
    sql_format = 'INSERT IGNORE INTO {database}.{table} ({fields}) VALUES ({values})'

    async def process(self, thing: Item):
        fields = []
        values = []
        for k, v in thing.items():
            fields.append(k)
            values.append(self.convert_item_value(v))
        sql = self.sql_format.format(database=self.database, table=self.table, fields=','.join(fields),
                                     values=','.join(values))
        await self.push_data(sql)
        return thing


class ItemMysqlUpdatePipeline(ItemBaseMysqlPipeline):
    sql_format = 'UPDATE {database}.{table} SET {pairs} WHERE {primary_key}={primary_value}'

    def __init__(self, primary_key: str, host: str, port: int, user: str, password: str, database: str, table: str,
                 charset: str='utf8'):
        super().__init__(host, port, user, password, database, table, charset=charset)
        self.primary_key = primary_key

    async def process(self, thing: Item) -> Item:
        pairs = []
        primary_value = None

        for k, v in thing.items():
            if k == self.primary_key:
                primary_value = self.convert_item_value(v)
            else:
                pairs.append('{:s}={:s}'.format(k, self.convert_item_value(v)))

        if primary_value is not None:
            sql = self.sql_format.format(database=self.database, table=self.table, pairs=','.join(pairs),
                                         primary_key=self.primary_key, primary_value=primary_value)
            await self.push_data(sql)
        return thing


class ItemMysqlInsertUpdatePipeline(ItemBaseMysqlPipeline):
    sql_format = 'INSERT INTO {database}.{table} ({fields}) VALUES ({values}) on duplicate key update {pairs}'

    def __init__(self, update_keys: List[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_keys = update_keys

    async def process(self, thing: Item) -> Item:
        fields = []
        values = []
        pairs = []
        for k, v in thing.items():
            v = self.convert_item_value(v)
            fields.append(k)
            values.append(v)
            if k in self.update_keys:
                pairs.append('{:s}={:s}'.format(k, v))
        sql = self.sql_format.format(database=self.database, table=self.table, fields=','.join(fields),
                                     values=','.join(values), pairs=','.join(pairs))
        await self.push_data(sql)
        return thing


class ItemBaseEmailPipeline(Pipeline):
    def __init__(self, account: str, password: str, server: str, port: int, recipients: List[str],
                 sender_name: str='AntNest.ItemEmailPipeline', tls: bool=False, starttls: bool=False):
        super().__init__()
        self.account = account
        self.password = password
        self.server = server
        self.port = port
        self.recipients = recipients
        self.sender_name = sender_name
        self.starttls = starttls
        self.tls = tls

    async def open_smtp(self) -> aiosmtplib.SMTP:
        smtp = aiosmtplib.SMTP()
        if self.starttls:
            await smtp.connect(self.server, self.port, use_tls=False)
            await smtp.starttls()
        else:
            await smtp.connect(self.server, self.port, use_tls=self.tls)
        await smtp.login(self.account, self.password)
        return smtp

    async def send(self, smtp: aiosmtplib.SMTP, title: str, content: str, attachments: Optional[List[IO]]=None):
        if attachments is None:
            msg = MIMEText(content)
        else:
            msg = MIMEMultipart()
            msg.attach(MIMEText(content))
            for f in attachments:
                att = MIMEText(f.read(), 'base64', 'utf-8')
                att["Content-Type"] = 'application/octet-stream'
                att["Content-Disposition"] = 'attachment; filename="{:s}"'.format(f.name)
                msg.attach(att)
        msg['From'] = '{:s} <{:s}>'.format(self.sender_name, self.account)
        msg['To'] = '<' + '> <'.join(self.recipients) + '>'
        msg['Subject'] = title
        await smtp.send_message(msg)


class ItemEmailPipeline(ItemBaseEmailPipeline):
    def __init__(self, title, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = []
        self.title = title

    def process(self, thing: Item) -> Item:
        self.items.append(thing)
        return thing

    async def on_spider_close(self) -> None:
        smtp = await self.open_smtp()
        await self.send(smtp, self.title, '\n'.join(item.__repr__() for item in self.items))
        smtp.close()


__all__ = [var for var in vars().keys() if 'Pipeline' in var]



"""CLI entry points"""
from typing import Type, Dict, List, Callable
import argparse
import inspect
import os
import sys
import asyncio
from importlib import import_module
from pkgutil import iter_modules
from traceback import format_exc
import webbrowser
import tempfile

from .ant import Ant
from .things import Response
from . import queen


__all__ = ['get_ants', 'run_ant', 'open_response_in_browser']


def get_ants(paths: List[str]) -> Dict[str, Type[Ant]]:
    """Get ant classes by package path"""
    modules = []
    results = {}
    # get all modules from packages and subpackages
    for path in paths:
        module = import_module(path)
        modules.append(module)
        if hasattr(module, '__path__'):
            package_path = []
            for _, name, ispkg in iter_modules(module.__path__):
                next_path = path + '.' + name
                if ispkg:
                    package_path.append(next_path)
                else:
                    modules.append(import_module(next_path))
            if len(package_path) > 0:
                results.update(get_ants(package_path))
    # get and sift ant class obj from modules
    for module in modules:
        for name, obj in inspect.getmembers(module):
            if isinstance(obj, type) and issubclass(obj, Ant) and obj is not Ant:
                results[module.__name__ + '.' + obj.__name__] = obj
    return results


async def run_ant(ant_cls: Type[Ant]):
    queen.init_loop(loop=asyncio.get_event_loop())
    ant = ant_cls()
    await ant.main()


def open_response_in_browser(response: Response, file_type: str='.html',
                             _open_browser_function: Callable[..., bool]=webbrowser.open) -> bool:
    fd, path = tempfile.mkstemp(file_type)
    os.write(fd, response._content)
    os.close(fd)
    return _open_browser_function('file://' + path)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--ant', help='ant name')
    parser.add_argument('-l', '--list', help='list ants', action='store_true')
    args = parser.parse_args()
    sys.path.append(os.getcwd())
    try:
        import settings
    except ImportError:
        print('Please create "settings.py" in project root directory or use the command in project root directory')
        exit(-1)
    try:
        ants = get_ants(settings.ANT_PACKAGES)
    except Exception as e:
        print('There is problem when finding and loading ants:')
        print(format_exc())
        exit(-1)
    if args.list:
        if len(ants) == 0:
            print('Can`t find any ant from ' + ','.join(settings.ANT_PACKAGES))
            exit(-1)
        else:
            print('\n'.join(ants.keys()))
    elif args.ant is not None:
        ant_name = args.ant
        if ant_name in ants:
            loop = getattr(settings, 'EVENT_LOOP', asyncio.get_event_loop())
            limit = getattr(settings, 'COROUTINE_CONCURRENT_LIMIT', None)
            queen.reset_concurrent_limit(limit)
            loop.run_until_complete(run_ant(ants[ant_name]))
        else:
            print('Can not find ant by the name "{:s}"'.format(ant_name))
            exit(-1)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
from setuptools import setup, find_packages


requires = ['aiohttp>=2.1.0', 'lxml>=3.7.0',
            'aiomysql>=0.0.11', 'PyMySQL>=0.7.11',
            'jpath>=1.5', 'aiosmtplib>=1.0.2', 'tenacity>=4.8.0',
            'pytest>=3.3.1', 'pytest-asyncio>=0.8.0', 'pytest-cov>=2.5.1']


setup(
    name="ant_nest",
    version="0.26",
    url='https://github.com/6ugman/ant_nest',
    description='A simple and clear Web Crawler framework build on python3.6+ with async',
    long_description=open('README.rst').read(),
    author='Bruce Wu',
    author_email='1wumingyu1@gmail.com',
    license='LGPL',
    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=('tests', 'tests.*')),
    install_requires=requires,
    entry_points={
        'console_scripts': ['ant_nest = ant_nest.cli:main']
    },
)



from typing import Any

from .ant import *
from .cli import *
from .things import *
from .pipelines import *
from . import queen
from .exceptions import *


def extract_value(_type: str, path: str, data: Any, extract_type: str=ItemExtractor.extract_with_take_first,
                  ignore_exception: bool=True,
                  default: Any=None) -> Any:
    try:
        return ItemExtractor.extract_value(_type, path, data, extract_type=extract_type)
    except Exception as e:
        if ignore_exception:
            return default
        else:
            raise e


def extract_value_by_xpath(path: str, data: Any, extract_type: str=ItemExtractor.extract_with_take_first,
                           ignore_exception: bool=True, default: Any=None) -> Any:
    return extract_value('xpath', path, data, extract_type=extract_type, ignore_exception=ignore_exception,
                         default=default)


def extract_value_by_jpath(path: str, data: Any, extract_type: str=ItemExtractor.extract_with_take_first,
                           ignore_exception: bool=True, default: Any=None) -> Any:
    return extract_value('jpath', path, data, extract_type=extract_type, ignore_exception=ignore_exception,
                         default=default)


def extract_value_by_regex(path: str, data: Any, extract_type: str=ItemExtractor.extract_with_take_first,
                           ignore_exception: bool=True, default: Any=None) -> Any:
    return extract_value('regex', path, data, extract_type=extract_type, ignore_exception=ignore_exception,
                         default=default)


__all__ = ['queen'] + ant.__all__ + cli.__all__ + pipelines.__all__ + exceptions.__all__ +\
          things.__all__ + ['extract_value_by_jpath', 'extract_value_by_regex', 'extract_value_by_xpath']
